# SPDX-License-Identifier: CECILL-C

const corpora_file = joinpath(@__DIR__, "corpora", "corpora.json")
const corpora_names = map(c -> c["name"], JSON.parsefile(corpora_file))

@kwdef struct SpeechDatasetInfos
    name::AbstractString = ""
    lang::Union{AbstractString, Vector{AbstractString}} = ""
    license::AbstractString = ""
    source::AbstractString = ""
    authors::Vector{AbstractString} = []
    description::AbstractString = ""
    subsets::Vector{AbstractString} = []
end

function SpeechDatasetInfos(infos::AbstractDict)
    kwargs = NamedTuple()
    for key in fieldnames(SpeechDatasetInfos)
        val = get(infos, String(key), nothing)
        # merge new (key=val) if key was found
        kwargs = !isnothing(val) ? (; kwargs..., key=>val) : kwargs
    end
    SpeechDatasetInfos(;kwargs...)
end

function SpeechDatasetInfos(name::AbstractString)
    corpora_infos = JSON.parsefile(corpora_file)
    infos = filter(x -> x["name"]==name, corpora_infos)[1]
    SpeechDatasetInfos(infos)
end


struct SpeechDataset <: MLUtils.AbstractDataContainer
    infos::SpeechDatasetInfos
    idxs::Vector{AbstractString}
    annotations::Dict{AbstractString, Annotation}
    recordings::Dict{AbstractString, Recording}
end

function SpeechDataset(infos::SpeechDatasetInfos, annotations::Dict{AbstractString, Annotation}, recordings::Dict{AbstractString, Recording})
    idxs = collect(keys(annotations))
    SpeechDataset(infos, idxs, annotations, recordings)
end

function SpeechDataset(infos::SpeechDatasetInfos, manifestroot::AbstractString, partition::AbstractString)
    partition_name = partition == "" ? "" : "-$(partition)"
    annot_path =  joinpath(manifestroot, "annotations$(partition_name).jsonl") 
    rec_path = joinpath(manifestroot, "recordings.jsonl")
    annotations = load_manifest(Annotation, annot_path)
    recordings = load_manifest(Recording, rec_path)
    SpeechDataset(infos, annotations, recordings)
end

Base.getindex(d::SpeechDataset, key::AbstractString) = d.recordings[key], d.annotations[key]
Base.getindex(d::SpeechDataset, idx::Integer) = getindex(d, d.idxs[idx])
# Fix1 -> partial function with fixed 1st argument
Base.getindex(d::SpeechDataset, idxs::AbstractVector) = map(Base.Fix1(getindex, d), idxs)

Base.length(d::SpeechDataset) = length(d.idxs)

function Base.filter(fn, d::SpeechDataset)
    fidxs = filter(d.idxs) do i
        fn((d.recordings[i], d.annotations[i]))
    end
    idset = Set(fidxs)

    fannotations = filter(d.annotations) do (k, v)
        k ∈ idset
    end

    frecs = filter(d.recordings) do (k, v)
        k ∈ idset
    end

    SpeechDataset(d.infos, fidxs, fannotations, frecs)
end


struct DatasetBuilder{name}
    kwargs::NamedTuple
end

DatasetBuilder(name::Symbol) = DatasetBuilder{name}() # still to be defined for each supported symbol

function declareBuilder(name::Symbol; kwargs::NamedTuple=(;))
    quotedname = Meta.quot(name)
    eval(Meta.parse("(::Type{DatasetBuilder{$quotedname}})() = DatasetBuilder{$quotedname}($kwargs)"))
end

get_nametype(name::String) = Symbol(replace(name, " "=>""))

function get_outfiles(subsets)
    outfiles = ["recordings.jsonl"]
    if isempty(subsets)
        push!(outfiles, "annotations.jsonl")
    else 
        annotationsfiles = ["annotations-$subset.jsonl" for subset in subsets]
        push!(outfiles, annotationsfiles...)
    end
    outfiles
end

delete(nt::NamedTuple{names}, key) where names = NamedTuple{filter(x -> x != key, names)}(nt)

"""
Extract metadata and paths from desired dataset.\n
Create the `outputdir` folder, with:
- `recordings.jsonl` containing each audio file path and associated metadata
- `annotations-<subset>.jsonl` containing each annotation and associated metadata
Return a SpeechDataset object.
# Arguments
- `name` is the name of the dataset. Supported names are $corpora_names.
- `inputdir` is the name of dataset directory. If the directory does not exists, it is created and the data is downloaded if possible. Not all datasets can be downloaded, for example proprietary datasets does not implements a download function.\n
- `outputdir` is the output directory for manifest files.
# Keyword Arguments
Common kwargs are
- `subset` is the part of the dataset to load (for example "train" or "test").\n
- `lang` is the ISO 639-3 code of the language.\n
Other kwargs can be available depending on the dataset
"""
function dataset(name::AbstractString, inputdir::AbstractString, outputdir::AbstractString; kwargs...)
    if ! (name ∈ corpora_names)
        throw(ArgumentError("Name $name is not supported, try one of $corpora_names."))
    end 
    nametype = get_nametype(name)
    dataset(DatasetBuilder(nametype), name, inputdir, outputdir; kwargs...)
end

function dataset(builder::DatasetBuilder, name::AbstractString, inputdir::AbstractString, outputdir::AbstractString; kwargs...)
    kwargs = values(kwargs)
    # download if possible and inputdir doesn't exists yet  
    downloadable = hasmethod(download, Tuple{typeof(builder), AbstractString})
    if downloadable && ! isdir(inputdir)
        download(builder, inputdir)
    end

    # load infos from dataset name
    infos = SpeechDatasetInfos(name)

    # get subset if available and check its value 
    if :subset in keys(kwargs)
        if ! (subset ∈ infos.subsets)
            throw(ArgumentError("Subset $subset is not supported, try one of $(infos.subsets) or omit this kwarg."))
        end 
        kwargs = delete(kwargs, :subset)
    end

    # Check lang value if provided
    if :lang in keys(kwargs) 
        if ! (lang in infos.lang)
            throw(ArgumentError("Language $lang is not supported, try one of $(infos.lang) or omit this kwarg."))
        end 
    end

    # prepare if outfiles are not present
    outfiles = get_outfiles(infos.subsets)
    if ! all(isfile.(joinpath.(outputdir, outfiles)))
        prepare(builder, inputdir, outputdir; kwargs...)
    end

    SpeechDataset(infos, outputdir, subset)
end


#TODO add summary() for SpeechDatasetInfos and SpeechDataset
#TODO function to list available kwargs